@@ -3,16 +3,21 @@ from __future__ import annotations
 import asyncio
 import logging
 import socket
+from asyncio import Lock
 from typing import Callable, Coroutine
 
 import async_timeout
 
-from roborock.api import RoborockClient
+from roborock.api import RoborockClient, SPECIAL_COMMANDS
 from roborock.exceptions import RoborockTimeout, CommandVacuumError
 from roborock.typing import RoborockCommand
 from roborock.util import get_running_loop_or_create_one
 
 secured_prefix = 199
+get_prefix = 119
+app_prefix = 135
+set_prefix = 151
+
 _LOGGER = logging.getLogger(__name__)
 
 
@@ -34,9 +39,10 @@ class RoborockLocalClient(RoborockClient):
     async def send_command(
             self, device_id: str, method: RoborockCommand, params: list = None
     ):
-        request_id, timestamp, payload = super()._get_payload(method, params)
+        secured = True if method in SPECIAL_COMMANDS else False
+        request_id, timestamp, payload = self._get_payload(method, params, secured)
         _LOGGER.debug(f"id={request_id} Requesting method {method} with {params}")
-        prefix = secured_prefix
+        prefix = secured_prefix if method in SPECIAL_COMMANDS else get_prefix
         protocol = 4
         msg = self._encode_msg(device_id, protocol, timestamp, payload, prefix)
         _LOGGER.debug(f"Requesting with prefix {prefix} and payload {payload}")
@@ -49,26 +55,38 @@ class RoborockLocalClient(RoborockClient):
         _LOGGER.debug(f"id={request_id} Response from {method}: {response}")
         return response
 
+class RoborockSocket(socket.socket):
+    _closed = None
+
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+
+    @property
+    def is_closed(self):
+        return self._closed
 
 class RoborockSocketListener:
     roborock_port = 58867
 
-    def __init__(self, ip: str, device_id: str, on_message: Callable[[str, bytes], Coroutine | None],
+    def __init__(self, ip: str, device_id: str, on_message: Callable[[str, bytes], Coroutine[bool] | bool],
                  timeout: float | int = 4):
         self.ip = ip
         self.device_id = device_id
-        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        self.socket = RoborockSocket(socket.AF_INET, socket.SOCK_STREAM)
         self.socket.setblocking(False)
         self.loop = get_running_loop_or_create_one()
         self.on_message = on_message
         self.timeout = timeout
         self.is_connected = False
+        self._lock = Lock()
 
     async def _main_coro(self):
-        while self.is_connected:
+        while not self.socket.is_closed:
             try:
                 message = await self.loop.sock_recv(self.socket, 4096)
-                await self.on_message(self.device_id, message)
+                accepted = await self.on_message(self.device_id, message)
+                if accepted:
+                    self._lock.release() if self._lock.locked() else None
             except Exception as e:
                 _LOGGER.exception(e)
                 self.is_connected = False
@@ -78,14 +96,16 @@ class RoborockSocketListener:
         async with async_timeout.timeout(self.timeout):
             await self.loop.sock_connect(self.socket, (self.ip, 58867))
             self.is_connected = True
-        asyncio.create_task(self._main_coro())
+        self.loop.create_task(self._main_coro())
 
     async def send_message(self, data: bytes, local_key: str):
         response = {}
         try:
             async with async_timeout.timeout(self.timeout):
+                await self._lock.acquire()
                 await self.loop.sock_sendall(self.socket, data)
         except (asyncio.TimeoutError, asyncio.CancelledError):
+            self._lock.release() if self._lock.locked() else None
             raise RoborockTimeout(
                 f"Timeout after {self.timeout} seconds waiting for response"
             ) from None
